/*
 * Copyright 2010 Martin Schreiber
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include "CGLTools.hpp"


void CGLTools::checkGLError_(const char *file, const char *function, int line)
{
	int nr = glGetError();
	const char *error = 0;
	switch(nr)
	{
		case GL_NO_ERROR:
			error = 0;
			break;

		case GL_INVALID_ENUM:
			error = "INVALID_ENUM";
			break;

		case GL_INVALID_VALUE:
			error = "INVALID_VALUE";
			break;
		
		case GL_INVALID_OPERATION:
			error = "INVALID_OPERATION";
			break;

		case GL_STACK_OVERFLOW:
			error = "INVALID_OVERFLOW";
			break;

		case GL_STACK_UNDERFLOW:
			error = "INVALID_UNDERFLOW";
			break;

		case GL_OUT_OF_MEMORY:
			error = "OUT_OF_MEMORY";
			break;

		case GL_TABLE_TOO_LARGE:
			error = "TABLE_TOO_LARGE";
			break;

		default:
			char unknown[100];
			sprintf(unknown, "UNKNOWN %i", nr);
			error = unknown;
	}

	if (error)
	{
		fprintf(stderr, "\033[0;40;31m");
		fprintf(stderr, "GL ERROR  in   %s(%i): %s() -> %s", file, line, function, error);
		fprintf(stderr, "\033[0m");
		fprintf(stderr, "\n");
	}

}

bool CGLTools::unFrustum(	float *matrix,
							float *left,
							float *right,
							float *bottom,
							float *top,
							float *near,
							float *far)
{
	float a = matrix[0];
	float c = matrix[4*2];
	float d = matrix[4*1 + 1];
	float e = matrix[4*2 + 1];
	float g = matrix[4*2 + 2];
	float h = matrix[4*3 + 2];
	
	if (matrix[4*2+3] != -1.0f)	// funny check
		return false;

	*near = (((g+1)/(g-1))*h - h)/2;
	*far = *near*(g-1)/(g+1);
	*left = (*near*2)/(((c+1)/(c-1)-1)*a);
	*right = *left*(c+1)/(c-1);
	*bottom = (*near*2)/(((e+1)/(e-1)-1)*d);
	*top = *bottom*(e+1)/(e-1);

	return true;
}
