/*
 * Copyright 2010 Martin Schreiber
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#ifndef CLATTICE_BOLTZMANN_HPP
#define CLATTICE_BOLTZMANN_HPP

#include <stdlib.h>
#include <iostream>
#include <unistd.h>
#include "libcl/CCL.hpp"
#include "libtools/CStopwatch.hpp"
#include "libtools/CEyeBall.hpp"
#include "CLbmOpenCl.hpp"
#include "CGui.hpp"
#include "CLbmVisualization.hpp"
#include <list>

#include "CPrimitives.hpp"
#include "CFont.hpp"

// simulation type
typedef float T;
//typedef double T;

template <typename T>
class CMain	:
       		public CGuiInterface
{
	CGui cGui;
	bool gui_quit;
	bool reset;
	CEyeBall<float> cEyeBall;
	CLbmVisualization<T> cLbmVisualization;
	int next_simulation_steps_count;
	CFont cFont;
	CLbmOpenCl<T> *cLbmPtr;

	int old_mouse_x, old_mouse_y;
	bool left_button_down;
	bool right_button_down;

	T vector_checksum;

	// amount of vectors to omit while visualization of velicities
	int visualization_increment;

	// true if debug mode is active
	bool debug_mode;

	// fullscreen mode active
	bool fullscreen;

	void outputDD(int dd_i)
	{
		std::cout << "DD " << dd_i << std::endl;
	//					int gcd = CMath<int>::gcd(cLbmPtr->domain_cells[0],wrap_max_line);
	//					if (wrap_max_line % gcd == 0)
	//						gcd = wrap_max_line;
		int wrap_max_line = 16;
		int gcd = wrap_max_line;
		cLbmPtr->debugDD(dd_i, gcd, cLbmPtr->domain_cells[0]*cLbmPtr->domain_cells[1]);
	}

#define DEBUG_DD	0
	void callback_key_down(int key, int mod, int scancode, int unicode)
	{
		size_t increment;
		size_t dd_debug_id = 0;

		switch(key)
		{
			case 'q':	case 'Q':
				gui_quit = true;
				break;

			case '0':
				if (next_simulation_steps_count < 0)
					next_simulation_steps_count = 0;

				increment = 1;
				next_simulation_steps_count += increment*increment;
				break;

			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
				if (debug_mode)
				{
					increment = (key - '1' + 1);
					outputDD(increment-1);
				}
				else
				{
					if (next_simulation_steps_count < 0)
						next_simulation_steps_count = 0;

					increment = (key - '1' + 1);
					next_simulation_steps_count += increment*increment;
				}
				break;

			case '!':	outputDD(9);	break;
			case '@':	outputDD(10);	break;
			case '#':	outputDD(11);	break;
			case '$':	outputDD(12);	break;
			case '%':	outputDD(13);	break;
			case '^':	outputDD(14);	break;
			case '&':	outputDD(15);	break;
			case '*':	outputDD(16);	break;
			case '(':	outputDD(17);	break;
			case ')':	outputDD(18);	break;

			case 'u':	// output dd information
				if (debug_mode)
				{
					std::cout << "DD " << dd_debug_id << std::endl;
//					int gcd = CMath<int>::gcd(cLbmPtr->domain_cells[0],wrap_max_line);
//					if (wrap_max_line % gcd == 0)
//						gcd = wrap_max_line;
					int wrap_max_line = 8;
					int gcd = wrap_max_line;
					cLbmPtr->debugDD(dd_debug_id, gcd, cLbmPtr->domain_cells[0]*cLbmPtr->domain_cells[1]);
				}
				break;

			case 'o':	
				if ((visualization_increment+1)/2 > cLbmPtr->domain_cells.min())
					break;
				visualization_increment++;
				break;

			case 'f':	// fullscreen
				fullscreen = !fullscreen;
				cGui.setVideoMode(fullscreen);
				break;

			case 'r':	// reset simulation
				reset = true;
				std::cout << "resetting..." << std::endl;
				break;

			case 'e':	// reload simulation code
				cLbmPtr->reload();
				break;

			case 'p':
				if (visualization_increment <= 1)
					break;
				visualization_increment--;
				break;

			case 'c':
				vector_checksum = cLbmPtr->getVelocityChecksum();
				break;

			case ' ':
				next_simulation_steps_count = -1;
				break;
		}

		// avoid warnings
		mod = 0;
		scancode = 0;
		unicode = 0;
	}

	void callback_key_up(int key, int mod, int scancode, int unicode)
	{
		key = 0;
		mod = 0;
		scancode = 0;
		unicode = 0;
	}

	void callback_quit()
	{
		gui_quit = true;
	}

	void callback_mouse_motion(int x, int y)
	{
		if (left_button_down)
		{
			cEyeBall.rotate(-(float)(old_mouse_x - x), -(float)(old_mouse_y - y), 0);
		}
		else if (right_button_down)
		{
			cLbmPtr->addDrivenCavityValue(-(float)(old_mouse_x - x));
		}

		old_mouse_x = x;
		old_mouse_y = y;
	}

	void callback_mouse_button_down(int button)
	{
		if (button == 1)
			left_button_down = true;
		else if (button == 3)
			right_button_down = true;
	}

	void callback_mouse_button_up(int button)
	{
		if (button == 1)
			left_button_down = false;
		else if (button == 3)
			right_button_down = false;
	}

	void callback_viewport_changed(int width, int height)
	{
		glViewport(0, 0, width, height);
	}

public:

	CMain()	:
		cGui(*this, "lattice boltzmann driven cavity test"),
		gui_quit(false),
		next_simulation_steps_count(-1),
		left_button_down(false),
		right_button_down(false),
		visualization_increment(-1),
		fullscreen(false)
	{
	}

	int run(
			bool p_debug_mode,
			CVector<3,int> domain_size,
			CVector<3,T> gravitation,
			T viscosity,
			size_t computation_kernel_count,
			int device_nr,
			bool run_with_gui,
			bool pause,
			T timestep,
			bool take_frame_screenshots,
			int loops,

			std::list<int> &p_lbm_opencl_number_of_work_items_list,		///< list with number of threads for each successively created kernel
			std::list<int> &p_lbm_opencl_number_of_registers_list		///< list with number of registers for each thread threads for each successively created kernel
	)		
	{
		if (loops < 0)
			loops = 100;

		debug_mode = p_debug_mode;

		if (debug_mode)
			std::cout << "domain size: " << domain_size << std::endl;

		if (pause)
			next_simulation_steps_count = 0;

		// load platform information
		if (debug_mode)	std::cout << "loading platforms" << std::endl;
		CCL::CPlatforms cPlatforms;
		cPlatforms.load();

		if (cPlatforms.platform_ids_count == 0)
		{
			std::cerr << "no platform found!" << std::endl;
			return -1;
		}

		int platform_id_nr = -1;
		for (size_t i = 0; i < cPlatforms.platform_ids_count; i++)
		{
			CCL::CPlatform cPlatform(cPlatforms.platform_ids[i]);
			cPlatform.loadPlatformInfo();

			if (platform_id_nr == -1)
				if (strcmp(cPlatform.profile, "FULL_PROFILE") == 0)
				{
					platform_id_nr = i;
					if (debug_mode)
						std::cout << "Using Platform " << (i+1) << " for computation" << std::endl;
				}

			if (debug_mode)
			{
				std::cout << "Platform " << (i) << ":" << std::endl;
				std::cout << "        Name: " << cPlatform.name << std::endl;
				std::cout << "     Profile: " << cPlatform.profile << std::endl;
				std::cout << "     Version: " << cPlatform.version << std::endl;
				std::cout << "      Vendor: " << cPlatform.vendor << std::endl;
				std::cout << "  Extensions: " << cPlatform.extensions << std::endl;
				std::cout << std::endl;
			}
		}

		if (platform_id_nr == -1)
		{
			std::cout << "no usable platform found" << std::endl;
			return -1;
		}

		CCL::CPlatform cPlatform(cPlatforms.platform_ids[platform_id_nr]);

		// load standard context for GPU devices
		if (debug_mode)	std::cout << "loading gpu context" << std::endl;
		CCL::CContext cContext(cPlatform, CL_DEVICE_TYPE_GPU);

		// load devices belonging to cContext
		if (debug_mode)	std::cout << "loading devices" << std::endl;
		CCL::CDevices cDevices(cContext);

		if (cDevices.size() == 0)
		{
			std::cerr << "no device found - aborting" << std::endl;
			return -1;
		}

		if (device_nr == -1)
		{
			// list available devices
			for (int i = 0; i < (int)cDevices.size(); i++)
			{
				CCL::CDeviceInfo cDeviceInfo(cDevices[i]);
				std::cout << "Device " << (i) << ":" << std::endl;
				std::cout << "        Name: " << cDeviceInfo.name << std::endl;
				std::cout << "     Profile: " << cDeviceInfo.profile << std::endl;
				std::cout << "     Version: " << cDeviceInfo.version << std::endl;
				std::cout << "      Vendor: " << cDeviceInfo.vendor << std::endl;
				std::cout << "  Extensions: " << cDeviceInfo.extensions << std::endl;
				std::cout << std::endl;
			}
			return -1;
		}

		if (device_nr < 0 || device_nr >= (int)cDevices.size())
		{
			std::cerr << "invalid device number - use option \"-d -1\" to list all devices" << std::endl;
			return -1;
		}

		CCL::CDevice &cDevice = cDevices[device_nr];

		// load information about first device - e.g. max_work_group_size
		if (debug_mode)	std::cout << "loading device information" << std::endl;
		CCL::CDeviceInfo cDeviceInfo(cDevice);

		if (debug_mode)
		{
			std::cout << "Device " << (device_nr) << ":" << std::endl;
			std::cout << "        Name: " << cDeviceInfo.name << std::endl;
			std::cout << "     Profile: " << cDeviceInfo.profile << std::endl;
			std::cout << "     Version: " << cDeviceInfo.version << std::endl;
			std::cout << "      Vendor: " << cDeviceInfo.vendor << std::endl;
			std::cout << "  Extensions: " << cDeviceInfo.extensions << std::endl;
			std::cout << std::endl;
		}

		// initialize queue
		if (debug_mode)	std::cout << "creating command queue" << std::endl;
		CCL::CCommandQueue cCommandQueue(cContext, cDevice);

		T domain_length = 0.05;


		if (visualization_increment == -1)
		{
			visualization_increment = domain_size.min()/8;
			if (visualization_increment < 1)
				visualization_increment = 1;
		}


		vector_checksum = 0;

		// approximate bandwidth
		double floats_per_cell = 0.0;

		// 19 density distribution which are read and written
		floats_per_cell += 19.0*2.0;

		// flag (obstacle, injection and fluid) is read
		floats_per_cell += 1.0;

		// velocity vector is also stored
		if (run_with_gui || debug_mode)
			floats_per_cell += 3;

		reset = true;

		if (run_with_gui)
		{
			cEyeBall.reset();
			cGui.create();
			cFont.setup();
			cFont.setFontSize(16);

			while (reset && !gui_quit)
			{
				reset = false;

				// INIT LATTICE BOLTZMANN!
				CLbmOpenCl<T> cLbm(	cCommandQueue, cContext, cDevice,
							domain_size,		// domain size
							domain_length,		// length of domain size in x direction
							gravitation,	// gravitation vector
							viscosity,
							computation_kernel_count,
							debug_mode,
							run_with_gui || debug_mode,
							false,
							timestep,

							p_lbm_opencl_number_of_work_items_list,
							p_lbm_opencl_number_of_registers_list
						);

				if (cLbm.error())
				{
					std::cout << cLbm.error.getString();
					std::cout << std::endl;
					std::cout << "sleeping for 1 second and resetting..." << std::endl;
					sleep(1);
					reset = true;
					continue;
				}

				cLbm.wait();

				cLbmPtr = &cLbm;

				CStopwatch cStopwatch;
				cStopwatch.start();

				double fps = 0;
				size_t last_simulation_step = 0;
				while (!gui_quit && !reset)
				{
					// simulation
					if (next_simulation_steps_count == -1)
					{
						cLbm.simulationStep();
					}
					else if (next_simulation_steps_count > 0)
					{
						cLbm.simulationStep();
						next_simulation_steps_count--;
					}
					cLbm.wait();

					cLbmVisualization.setup(cLbm);

					// clear background
					glClearColor(0, 0, 0, 1);
					glClear(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT);

						glMatrixMode(GL_PROJECTION);
						glLoadIdentity();
						glFrustum(-0.5*cGui.aspect, 0.5*cGui.aspect, -0.5, 0.5, 1, 10);

						glMatrixMode(GL_MODELVIEW);
						glLoadIdentity();

							glPushMatrix();
							glTranslatef(0, 0, -2);

							glMultMatrixf((GLfloat*)cEyeBall.rotationMatrix);

							cLbmVisualization.render(visualization_increment);

							glColor3f(1, 0, 0);
							glScalef((float)domain_size[0]/(float)domain_size.max(), (float)domain_size[1]/(float)domain_size.max(), (float)domain_size[2]/(float)domain_size.max());
							CPrimitives::box();

							glPopMatrix();

						int posy = 5;
						glColor3f(1,1,1);

						cFont.printfxy(5, posy, "simulation steps: %i", cLbm.simulation_step_counter);	posy += 20;
						cFont.printfxy(5, posy, "remaining steps: %i", next_simulation_steps_count);	posy += 20;

						if (cStopwatch.getTime() > 1.0)
						{
							cStopwatch.stop();
							fps = (float)(cLbm.simulation_step_counter-last_simulation_step)/(float)cStopwatch.time;
							last_simulation_step = cLbm.simulation_step_counter;

							cStopwatch.reset();
							cStopwatch.start();
						}

						cFont.printfxy(5, posy, "fps: %f", fps);	posy += 20;
						double mlups = (fps*(double)cLbm.domain_cells.elements())*(double)0.000001;
						cFont.printfxy(5, posy, "MLUPS: %f", mlups);	posy += 20;
						// approximate bandwidth
						cFont.printfxy(5, posy, "Bandwidth: %f MB/s (RW, bidirectional)", mlups*floats_per_cell);	posy += 20;
						cFont.printfxy(5, posy, "inv_tau: %f", cLbm.inv_tau);	posy += 20;
						cFont.printfxy(5, posy, "d_timestep: %f", cLbm.d_timestep);	posy += 20;
						cFont.printfxy(5, posy, "visualization increment: %i", visualization_increment);	posy += 20;
						cFont.printfxy(5, posy, "driven cavity velocity: %f", cLbm.drivenCavityVelocity[0]);	posy += 20;

						if (debug_mode)
						{
							if (cLbm.simulation_step_counter <= 10)
								vector_checksum = cLbm.getVelocityChecksum();
						}
						cFont.printfxy(5, posy, "vector checksum: %.8f", vector_checksum*1000.0f);	posy += 20;

					cGui.swapBuffer();
					cGui.eventLoop();
				}
			}

			std::cout << std::endl;
			std::cout << "exit" << std::endl;
		}
		else
		{
			// INIT LATTICE BOLTZMANN!
			CLbmOpenCl<T> cLbm(	cCommandQueue, cContext, cDevice,
						domain_size,		// domain size
						domain_length,		// length of domain size in x direction
						gravitation,	// gravitation vector
						viscosity,
						computation_kernel_count,
						debug_mode,
						run_with_gui || debug_mode,
						false,
						timestep,

						p_lbm_opencl_number_of_work_items_list,
						p_lbm_opencl_number_of_registers_list
					);

			if (cLbm.error())
			{
				std::cout << cLbm.error.getString();
				return -1;
			}

			cLbm.wait();

			cLbmPtr = &cLbm;

			CStopwatch cStopwatch;

			for (int i = 0; i < loops/10; i++)
			{
				// simulation
				cLbm.simulationStep();
				std::cout << "." << std::flush;
			}
			std::cout << "|" << std::flush;
			cLbm.wait();

			cStopwatch.start();
			for (int i = 0; i < loops; i++)
			{
				// simulation
				cLbm.simulationStep();
				std::cout << "." << std::flush;
			}
			cLbm.wait();
			cStopwatch.stop();

			std::cout << std::endl;

			if (domain_size.elements() <= 512)
				if (debug_mode)
					cLbm.debug_print();

			std::cout << std::endl;

			std::cout << "Cube: " << domain_size << std::endl;
			std::cout << "Seconds: " << cStopwatch.time << std::endl;
			double fps = (((double)loops) / cStopwatch.time);
			std::cout << "FPS: " << fps << std::endl;

			double mlups = ((double)fps*(double)cLbm.domain_cells.elements())*(double)0.000001;
			std::cout << "MLUPS: " << mlups << std::endl;

			std::cout << "Bandwidth: " << (mlups*floats_per_cell*(double)sizeof(T)) << " MB/s (RW, bidirectional)" << std::endl;

			streamsize ss = std::cout.precision();
			std::cout.precision(8);
			std::cout.setf(ios::fixed,ios::floatfield);

			if (debug_mode)
			{
				// The velocity checksum is only stored in debug mode!
				vector_checksum = cLbm.getVelocityChecksum();
				std::cout << "Checksum: " << (vector_checksum*1000.0f) << std::endl;
			}

			std::cout.precision(ss);
			std::cout << std::resetiosflags(ios::fixed);

			std::cout << std::endl;
			std::cout << "exit" << std::endl;
		}

		return EXIT_SUCCESS;
	}
};

#endif
